@@ -839,6 +839,7 @@ bulk88\100hotmail.com                   bulk88
 
 chad.granum\100dreamhost.com            exodist7\100gmail.com
 choroba\100cpan.org                     choroba\100weed.(none)
++                                       choroba\100matfyz.cz
 chromatic\100wgz.org                    chromatic\100rmci.net
 ckuskie\100cadence.com                  colink\100perldreamer.com
 claes\100surfar.nu                      claes\100versed.se
@@ -936,6 +937,8 @@ jbuehler\100hekimian.com                jhpb\100hekimian.com
 jcromie\100cpan.org                      jcromie\100100divsol.com
 +                                       jim.cromie\100gmail.com
 jd\100cpanel.net                        lightsey\100debian.org
++                                       john\10004755.net
++                                       john\100nixnuts.net
 jdhedden\100cpan.org                    jerry\100hedden.us
 +                                       jdhedden\1001979.usna.com
 +                                       jdhedden\100gmail.com
@@ -1106,13 +1109,14 @@ khw\100cpan.org                         khw\100karl.(none)
 radu\100netsoft.ro                      rgreab\100fx.ro
 rajagopa\100pauline.schrodinger.com     rajagopa\100schrodinger.com
 raphael.manfredi\100pobox.com           raphael_manfredi\100grenoble.hp.com
-module@renee-baecker.de                 renee.baecker\100smart-websolutions.de
+module\100renee-baecker.de              renee.baecker\100smart-websolutions.de
 +                                       reneeb\100reneeb-desktop.(none)
-+                                       github@renee-baecker.de
++                                       github\100renee-baecker.de
 +                                       otrs\100ubuntu.(none)
 +                                       perl\100renee-baecker.de
 +                                       reb\100perl-services.de
 +                                       info\100perl-services.de
+rich+perl\100hyphen-dash-hyphen.info    richardleach\100users.noreply.github.com
 richard.foley\100rfi.net                richard.foley\100t-online.de
 +                                       richard.foley\100ubs.com
 +                                       richard.foley\100ubsw.com
