@@ -140,6 +140,7 @@ if (! $define{NO_LOCALE}) {
     }
 }
 
+# https://en.wikipedia.org/wiki/Microsoft_Visual_C%2B%2B#Internal_version_numbering
 my $cctype = $ARGS{CCTYPE} =~ s/MSVC//r;
 if (! $define{HAS_SETLOCALE} && $define{HAS_POSIX_2008_LOCALE}) {
     $define{USE_POSIX_2008_LOCALE} = 1;
@@ -398,6 +399,7 @@ unless ($define{'USE_ITHREADS'}) {
 		    PL_regex_pad
 		    PL_regex_padav
 		    PL_dollarzero_mutex
+		    PL_env_mutex
 		    PL_hints_mutex
 		    PL_locale_mutex
 		    PL_lc_numeric_mutex
@@ -753,7 +755,7 @@ unless ($define{'USE_QUADMATH'}) {
     foreach (@$embed) {
 	my ($flags, $retval, $func, @args) = @$_;
 	next unless $func;
-	if (($flags =~ /[AX]/ && $flags !~ $excludedre)
+	if (($flags =~ /[AXC]/ && $flags !~ $excludedre)
             || (!$define{'NO_MATHOMS'} && $flags =~ /b/))
         {
 	    # public API, so export
@@ -1306,7 +1308,7 @@ if ($ARGS{PLATFORM} eq 'win32') {
     print "LIBRARY $dll\n";
     # The DESCRIPTION module definition file statement is not supported
     # by VC7 onwards.
-    if ($ARGS{CCTYPE} =~ /^(?:MSVC60|GCC)$/) {
+    if ($ARGS{CCTYPE} eq 'GCC') {
 	print "DESCRIPTION 'Perl interpreter'\n";
     }
     print "EXPORTS\n";
