@@ -703,8 +703,9 @@ struct opslot {
 struct opslab {
     OPSLAB *	opslab_next;		/* next slab */
     OPSLAB *	opslab_head;		/* first slab in chain */
-    OP *	opslab_freed;		/* chain of freed ops (head only)*/
+    OP **	opslab_freed;		/* array of sized chains of freed ops (head only)*/
     size_t	opslab_refcnt;		/* number of ops (head slab only) */
+    U16         opslab_freed_size;      /* allocated size of opslab_freed */
     U16		opslab_size;		/* size of slab in pointers,
                                            including header */
     U16         opslab_free_space;	/* space available in this slab
